from logging import getLogger
from typing import Final

from certbot._internal.cert_manager import _describe_certs
from certbot._internal.storage import renewal_file_for_certname
from certbot.display.util import notify

from ..utils.config import AdoptedCertConfig
from ..utils.tlsa.updater import TlsaUpdater
from .base import HandlerBase

_LOGGER: Final = getLogger(name=__name__)


class ListAdoptablesHandler(HandlerBase):
    'Class for listing adoptable certificates.'

    def _process(self) -> int:
        'Show the adoptable certificate paths.'
        adoptable_certs = [
            self.all_certs.parsed[x] for x in self.all_certs.parsed
            if x not in self.all_certs.already_adopted]
        _describe_certs(
            config=self._certbot_config, parsed_certs=adoptable_certs,
            parse_failures=self.all_certs.parse_failures)
        return 0


class AdoptHandler(HandlerBase):
    'Adopting a previously deployed certificate by certbot.'

    def _get_certbot_args(self) -> list[str]:
        """
        Return command line arguments for certbot from the `daneupdate`
        passed arguments.
        """
        return super()._get_certbot_args()

    def _get_adoptable_configs(self) -> list[str]:
        """
        Return the list of adoptable configs (absolute paths) while
        doing preliminary checks and raising upon problems.
        """
        configs_to_adopt = set(
            renewal_file_for_certname(config=self._certbot_config, certname=x)
            for x in self._args.hostnames)
        if self.all_certs.already_adopted:
            str_already_adopted = \
                '\n'.join(f'- {x}' for x in self.all_certs.already_adopted)
            notify(
                msg='The following configurations have already been ' +
                f'adopted, not adopting:\n{str_already_adopted}')
        return [
            x for x in configs_to_adopt
            if x not in self.all_certs.already_adopted]

    def _process(self) -> int:
        'Internal processing, after certbot setup.'
        for conf_file in self._get_adoptable_configs():
            cert = self.all_certs.parsed[conf_file]
            notify(msg=f'Adopting {conf_file}')
            # Add with defaults
            adopted = AdoptedCertConfig.get_empty(configpath=conf_file)
            self._config.runtime.adopted[conf_file] = adopted
            if not self._args.skip_tlsa_update:
                tlsa_updater = TlsaUpdater(
                    verbose=self._args.verbose, dry_run=self._args.dry_run,
                    config=self._config, cert=cert, certconfig_path=conf_file)
                tlsa_updater.process()
            self._config.runtime.save(path=self._config.paths.runtimeyaml)
            notify(
                msg=f'Adopted {conf_file}. For manual changes, please check ' +
                str(self._config.paths.runtimeyaml))
        return 0
