from __future__ import annotations

from abc import ABCMeta, abstractmethod
from argparse import Namespace
from functools import cached_property
from logging import getLogger
from traceback import format_exc
from types import MappingProxyType
from typing import NamedTuple

from certbot._internal.cli import prepare_and_parse_args
from certbot._internal.display.obj import set_display
from certbot._internal.main import make_displayer
from certbot._internal.plugins.disco import PluginsRegistry
from certbot._internal.storage import renewal_conf_files
from certbot.crypto_util import verify_renewable_cert
from typing_extensions import Self

from ..utils.cert import DaneCert
from ..utils.config import Configuration

_LOGGER = getLogger(name=__name__)


class _ParsedCertsResult(NamedTuple):
    'Result of a `_get_parsed_certs()` call.'
    # A `MappingProxyType` is basically a read-only `dict`
    # config path -> DaneCert
    parsed: MappingProxyType[str, DaneCert]
    # lineage name -> config path
    lineages: MappingProxyType[str, str]
    parse_failures: tuple[str, ...]
    # set of config path
    already_adopted: frozenset[str]


class HandlerBase(object, metaclass=ABCMeta):
    'Base class for command handlers.'

    def __init__(self, args: Namespace, config: Configuration):
        self._args = args
        self._config = config
        self._plugins = PluginsRegistry.find_all()
        self._certbot_config = prepare_and_parse_args(
            plugins=self._plugins, args=self._get_certbot_args())

    def _get_certbot_args(self) -> list[str]:
        """
        Return command line arguments for certbot from the `daneupdate`
        passed arguments.
        """
        return []

    def _get_adopted_configpaths(self, available: set[str]) -> frozenset[str]:
        'Return the already adopted paths in the passed list of paths.'
        return frozenset(self._config.runtime.adopted).intersection(available)

    @cached_property
    def all_certs(self) -> _ParsedCertsResult:
        """
        Return a structured tuple of:
        - list of certbot certificates from configs created by certbot
        - paths of the configs that failed to parse
        - the daneupdate adopted configurations from the above ones

        See `certbot._internal.cert_manager.certificates()`.
        """
        parsed_certs = dict[str, DaneCert]()
        parse_failures = tuple[str, ...]()
        lineages = dict[str, str]()
        for renewal_file in renewal_conf_files(config=self._certbot_config):
            try:
                renewal_candidate = DaneCert.from_configdata(
                    config_filename=renewal_file,
                    cli_config=self._certbot_config)
                verify_renewable_cert(renewable_cert=renewal_candidate)
                parsed_certs[renewal_file] = renewal_candidate
                lineages[renewal_candidate.lineagename] = renewal_file
            except Exception as e:  # pylint: disable=broad-except
                _LOGGER.warning(msg=(
                    f'Renewal configuration file {renewal_file} '
                    f'produced an unexpected error: {e}. Skipping.'
                ))
                _LOGGER.debug(msg=f'Traceback was:\n{format_exc()}')
                parse_failures += (renewal_file,)
        already_adopted = self._get_adopted_configpaths(
            available=set(parsed_certs))
        return _ParsedCertsResult(
            parsed=MappingProxyType(parsed_certs),
            lineages=MappingProxyType(lineages),
            parse_failures=parse_failures, already_adopted=already_adopted)

    @abstractmethod
    def _process(self) -> int:
        'Implement this in subclasses.'
        raise NotImplementedError()

    def process(self) -> int:
        'Start processing.'
        with make_displayer(config=self._certbot_config) as displayer:
            set_display(display=displayer)
            return self._process()

    @classmethod
    def init_and_process(
        cls: type[Self], args: Namespace, config: Configuration
    ) -> int:
        obj = cls(args=args, config=config)
        return obj.process()
