from datetime import datetime, timedelta, timezone
from functools import cached_property
from logging import getLogger
from os.path import normpath
from typing import Final, Optional

from certbot._internal.cli.cli_utils import flag_default
from certbot._internal.client import Client
from certbot._internal.display.util import summarize_domain_list
from certbot._internal.hooks import (
    post_hook, pre_hook, renew_hook, run_saved_post_hooks)
from certbot._internal.main import _init_le_client, renew_cert
from certbot._internal.plugins.disco import PluginsRegistry
from certbot._internal.plugins.selection import choose_configurator_plugins
from certbot._internal.renewal import (
    _avoid_invalidating_lineage, _avoid_reuse_key_conflicts,
    _update_renewal_params_from_key, should_renew)
from certbot._internal.updater import (
    run_generic_updaters, run_renewal_deployer)
from certbot.display.util import notify
from certbot.interfaces import Authenticator, Installer
from certbot.util import Key
from ktools.pathlib import Path

from ..utils.cert import DaneCert
from ..utils.tlsa.updater import TlsaUpdater
from .base import HandlerBase

HELP_RENEW_DESC: Final = """\
Renew adopted and non-adopted certificates. Use this instead of \
certbot's own renew command.\
"""
HELP_RENEW_DRYRUN: Final = """\
Perform a test run against the Let's Encrypt staging server, obtaining \
test (invalid) certificates but not saving them to disk.\
"""
_LOGGER = getLogger(name=__name__)


class _CertbotLogicEmulator(object):
    'Emulating certbot renewal operations, for **one** `DaneCert`.'

    def __init__(self, cert: DaneCert, plugins: PluginsRegistry):
        self.cert = cert
        self._plugins = plugins

    @cached_property
    def _is_privkey_reused(self) -> bool:
        return self.cert.cli_config.reuse_key and \
            not self.cert.cli_config.new_key

    @cached_property
    def _installer(self) -> Optional[Installer]:
        installer, self._authenticator = choose_configurator_plugins(
            config=self.cert.cli_config, plugins=self._plugins,
            verb='certonly')
        return installer

    @cached_property
    def _authenticator(self) -> Optional[Authenticator]:
        self._installer, authenticator = choose_configurator_plugins(
            config=self.cert.cli_config, plugins=self._plugins,
            verb='certonly')
        return authenticator

    @cached_property
    def _le_client(self) -> Client:
        'An initialized LetsEncrypt HTTP client.'
        return _init_le_client(
            config=self.cert.cli_config, authenticator=self._authenticator,
            installer=self._installer)

    def _obtain_new_certificate(self) -> tuple[bytes, bytes, Key]:
        """
        The private key is the existing cert private key if `reuse_key`
        is set. Otherwise, generate a fresh private key by passing
        `None`.
        """
        if self._is_privkey_reused:
            new_priv_strpath = normpath(self.cert.privkey)  # type: ignore
            _update_renewal_params_from_key(
                key_path=new_priv_strpath, config=self.cert.cli_config)
        else:
            new_priv_strpath = None
        new_cert, new_chain, new_privkey, _ = \
            self._le_client.obtain_certificate(
                domains=self.cert.common_names, old_keypath=new_priv_strpath)
        return new_cert, new_chain, new_privkey

    def _emulate_certbot_renewal_renew_cert(self) -> None:
        """
        Use an adjusted renewal logic of:
        :func:`certbot._internal.renewal.renew_cert`
        """
        renewal_params: dict = \
            self.cert.configuration['renewalparams']  # type: ignore
        original_server = \
            renewal_params.get('server', flag_default('server'))
        _avoid_invalidating_lineage(
            config=self.cert.cli_config, lineage=self.cert,
            original_server=original_server)
        _avoid_reuse_key_conflicts(
            config=self.cert.cli_config, lineage=self.cert)
        new_cert, new_chain, new_privkey = self._obtain_new_certificate()
        if self.cert.cli_config.dry_run:
            cert_dirpath = Path(self.cert.cert).parent  # type: ignore
            _LOGGER.debug(msg=(
                f'Dry run: skipping updating lineage at {cert_dirpath}'))
            return
        self.cert.save_to_upcoming(
            new_cert=new_cert, new_chain=new_chain,
            new_privkey=new_privkey, is_privkey_reused=self._is_privkey_reused)

    def run_renew_adopted_logic(self) -> None:
        """
        Do a renewal using certbot's own internals, when
        `should_renew()`. This should put newly fetched certificates
        into the upcoming directory.

        Uses code paths defined in here:
        `certbot._internal.main.renew_cert()`
        """
        notify(msg='{action} for {domains}'.format(
            action='Simulating renewal of an existing certificate'
            if self.cert.cli_config.dry_run else
            'Renewing an existing certificate',
            domains=summarize_domain_list(self.cert.common_names)))
        self._emulate_certbot_renewal_renew_cert()

    def run_renew_nonadopted_logic(self) -> tuple[list[str], list[str]]:
        """
        Run a logic of renewing a certificate that\'s not adopted.
        Uses code paths defined in here:
        `certbot._internal.renewal.handle_renewal_request()`
        Hooks are handled by the underlying logic, that is in
        `certbot._internal.main._get_and_save_cert()`
        """
        renewed_domains = []
        failed_domains = []
        if not should_renew(config=self.cert.cli_config, lineage=self.cert):
            return renewed_domains, failed_domains
        self.cert.cli_config.certname = self.cert.lineagename
        # Empty config domains so handle_renewal_request doesn't bail
        # out
        config = self.cert.cli_config
        self.cert.ensure_deployed()
        renew_cert(config=config, plugins=self._plugins, lineage=self.cert)
        renewed_domains.extend(self.cert.common_names)
        run_generic_updaters(
            config=config, lineage=self.cert, plugins=self._plugins)

        # return handle_renewal_request(config=self.cert.cli_config)
        return renewed_domains, failed_domains

    def run_installer_logic(self) -> list[str]:
        """
        If a new certificate was put in place, run the installers/hooks
        for it.

        See `certbot._internal.main.renew_cert()`.

        :returns: The list of deployed cert's domains.
        """
        config = self.cert.cli_config
        pre_hook(config=config)
        self.cert.set_upcoming_as_latest()
        renew_hook(
            config=config, domains=self.cert.common_names,
            lineage_path=self.cert.live_dir)
        post_hook(config=config, renewed_domains=self.cert.common_names)
        if not self._installer:
            return []
        # In case of a newly-put-in-place certificate, reload server to
        # pick it up.
        run_renewal_deployer(
            config=config, lineage=self.cert, installer=self._installer)
        notify(msg=(
            f'Reloading {self._installer} server after certificate renewal'))
        self._installer.restart()
        return self.cert.common_names


class RenewHandler(HandlerBase):
    """
    Handle renewal of certificates, completely replacing certbot's
    `renew` command.
    """

    def _get_certbot_args(self) -> list[str]:
        """
        Return command line arguments for certbot from the `daneupdate`
        passed arguments.
        """
        args = super()._get_certbot_args()
        # The `renew` verb is needed to change hook executing logic
        args.append('renew')
        if self._args.test_cert:
            args.append('--test-cert')
        if self._args.force_renewal:
            args.append('--force-renewal')
        if self._args.dry_run:
            args.append('--dry-run')
        if self._args.verbose:
            args.append('--verbose')
        return args

    def _put_in_place_when_ttl_passed(
            self, certconfig_path: str, cert: DaneCert) -> list[str]:
        """
        If the upcoming cert's `mtime` has passed the TLSA record's TTL,
        put it in place.
        :returns: The certificate's domain names if it was put in place,
        for renewal hooks to run.
        """
        cert_config = self._config.runtime.adopted[certconfig_path]
        if cert_config.max_ttl == 0:
            cert_config.max_ttl = self._config.defaults.tlsa.ttl
        utcnow = datetime.now(tz=timezone.utc).timestamp()
        delta = cert.latest_upcoming_mtime + cert_config.max_ttl - utcnow
        tdelta = timedelta(seconds=delta)
        if delta > 0:
            # No putting in place necessary yet
            notify(msg=(
                'Certificate available in upcoming directory, waiting for TTL '
                f'expiration ({tdelta} remaining).'))
            return []
        notify(msg='TTL expired: moving upcoming certificates in place.')
        certbot_emu = _CertbotLogicEmulator(cert=cert, plugins=self._plugins)
        return certbot_emu.run_installer_logic()

    def _renew_one_adopted(
            self, certconfig_path: str, cert: DaneCert) -> list[str]:
        """
        Handle renewal/update of an adopted certificate.
        :returns: The list of renewed domains in case a previously
        fetched certificate was put in place, so renewal hooks can be
        run on them. If a certificate was newly fetched and put into the
        upcoming directory, its domain name is not returned.
        """
        tlsa_updater = TlsaUpdater(
            verbose=self._args.verbose, dry_run=self._args.dry_run,
            config=self._config, cert=cert, certconfig_path=certconfig_path)
        if cert.all_upcomings_in_place:
            domains = self._put_in_place_when_ttl_passed(
                certconfig_path=certconfig_path, cert=cert)
            if domains:
                # In case time_to_reload < 0, we need to update TLSAs
                tlsa_updater.process()
            return domains
        cert.cli_config.certname = cert.lineagename
        if not should_renew(config=cert.cli_config, lineage=cert):
            tlsa_updater.process()
            return []
        certbot_emu = _CertbotLogicEmulator(cert=cert, plugins=self._plugins)
        certbot_emu.run_renew_adopted_logic()
        tlsa_updater.process()
        return []

    def _start_renewing(
            self, to_renew: dict[str, DaneCert]) -> int:
        'Start renewing the chosen certificates.'
        renewed_domains = list[str]()
        failed_domains = list[str]()
        for certconfig_path, cert in to_renew.items():
            notify(msg=f'===== Daneupdate renewal: Checking {certconfig_path}')
            if certconfig_path in self.all_certs.already_adopted:
                renewed_domains.extend(self._renew_one_adopted(
                    certconfig_path=certconfig_path, cert=cert))
                # Add empty line
                notify(msg='')
                continue
            certbot_emu = _CertbotLogicEmulator(
                cert=cert, plugins=self._plugins)
            renewed, failed = certbot_emu.run_renew_nonadopted_logic()
            renewed_domains.extend(renewed)
            failed_domains.extend(failed)
            # Add empty line
            notify(msg='')
        run_saved_post_hooks(
            renewed_domains=renewed_domains, failed_domains=failed_domains)
        return 0

    def _process(self) -> int:
        'Renew certificates one by one.'
        if self._args.hostnames:
            # Renew specific hostnames
            hostname_set = set[str](self._args.hostnames)
            nonexistent_lineages = hostname_set - set(self.all_certs.lineages)
            if nonexistent_lineages:
                domains = summarize_domain_list(
                    domains=sorted(nonexistent_lineages))
                notify(
                    msg='The following passed hostnames (lineages) are ' +
                    f'not known by certbot: {domains}')
                return 1
            to_renew = {
                self.all_certs.lineages[x]:
                self.all_certs.parsed[self.all_certs.lineages[x]]
                for x in self._args.hostnames
            }
        else:
            # Renew ALL the hostnames
            to_renew = self.all_certs.parsed.copy()
        return self._start_renewing(to_renew=to_renew)
