from __future__ import annotations

from copy import deepcopy
from datetime import datetime, timezone
from functools import cached_property, lru_cache
from hashlib import sha256, sha512
from logging import getLogger
from typing import Literal, Mapping, Optional

from certbot._internal.renewal import reconstitute
from certbot._internal.storage import (
    ALL_FOUR, BASE_PRIVKEY_MODE, RenewableCert, config_with_defaults,
    update_configuration)
from certbot.compat.filesystem import (
    compute_private_key_mode, copy_ownership_and_apply_mode)
from certbot.configuration import NamespaceConfig
from certbot.crypto_util import verify_renewable_cert
from certbot.util import Key
from cryptography.hazmat.primitives._serialization import (
    Encoding, PublicFormat)
from cryptography.x509.base import Certificate, load_pem_x509_certificate
from ktools.pathlib import Path

from .config_types.common import TlsaMatchingType, TlsaSelector, TlsaUsage
from .tlsa.record_type import TlsaResourceRecord

KindType = Literal['cert', 'privkey', 'chain', 'fullchain']
_TargetDictType = dict[KindType, Path]
_LOGGER = getLogger(name=__name__)


def get_tlsa_bytes(
    cert: Certificate, selector: TlsaSelector,
    matching_type: TlsaMatchingType
) -> bytes:
    'Return a generated `bytes` data of the requested variant.'
    if selector == TlsaSelector.ENTIRE:
        data = cert.public_bytes(encoding=Encoding.DER)
    elif selector == TlsaSelector.PUBLICKEY:
        data = cert.public_key().public_bytes(
            encoding=Encoding.DER,
            format=PublicFormat.SubjectPublicKeyInfo)
    if matching_type == TlsaMatchingType.ENTIRE:
        return data
    elif matching_type == TlsaMatchingType.SHA256:
        return sha256(string=data).digest()
    elif matching_type == TlsaMatchingType.SHA512:
        return sha512(string=data).digest()


class DaneCert(RenewableCert):
    """
    Wrapping and extending a `RenewableCert` to have added methods &
    functionality.
    """
    _UPCOMING_DIRNAME = 'upcoming'
    __CACHED_PROPERTIES = (
        'latest_upcoming_mtime', 'common_names', 'all_upcomings_in_place',
        'latest_installed_mtime', '_DaneCert__upcoming_pubcert',
        '_DaneCert__second_latest_pubcert', '_DaneCert__latest_pubcert')

    def __init__(self):
        'Placeholder to override parent\'s `__init__`'

    def __reset_cache(self):
        self.configuration
        'Resed cached properties.'
        for name in self.__CACHED_PROPERTIES:
            if name in self.__dict__:
                delattr(self, name)
        self.get_tlsa_recordtypes.cache_clear()
        self.__is_oldcert_tlsa_needed.cache_clear()

    @staticmethod
    def wrap_renewablecert(cert: RenewableCert) -> DaneCert:
        'Return a `DaneCert` from a `RenewableCert`.'
        verify_renewable_cert(renewable_cert=cert)
        result = DaneCert()
        result.__dict__ = cert.__dict__
        return result

    @staticmethod
    def from_configdata(
            config_filename: str, cli_config: NamespaceConfig) -> DaneCert:
        'Return a `DaneCert` from configuration data.'
        cert = reconstitute(
            config=deepcopy(x=cli_config), full_path=config_filename)
        if not cert:
            raise RuntimeError(
                f'Error loading {config_filename}, please check the logs in ' +
                f'{cli_config.logs_dir}')
        return DaneCert.wrap_renewablecert(cert=cert)

    @cached_property
    def upcoming_dir(self) -> Path:
        return Path(self.archive_dir).joinpath(self._UPCOMING_DIRNAME)

    def get_upcoming_path(self, kind: KindType, version: int):
        'Return the upcoming version `Path` of the requested file.'
        original_path = Path(self.version(kind=kind, version=version))
        return original_path.parent.joinpath(
            self._UPCOMING_DIRNAME).joinpath(original_path.name)

    @cached_property
    def latest_installed_mtime(self) -> float:
        'Return the latest mtime of the installed certificate.'
        version = self.latest_common_version()
        self.cert
        targets: _TargetDictType = {
            kind: Path(self.version(kind=kind, version=version))
            for kind in ALL_FOUR}
        return max((
            Path.stat(x).st_mtime for x in targets.values()
        ), default=0)

    @cached_property
    def all_upcomings_in_place(self) -> _TargetDictType:
        'Return all of the upcoming certificate `Path`s.'
        self.upcoming_dir.mkdir(exist_ok=True)
        next_version = self.next_free_version()
        all_upcoming: _TargetDictType = {
            kind: self.get_upcoming_path(kind=kind, version=next_version)
            for kind in ALL_FOUR}
        if all(Path.is_file(x) for x in all_upcoming.values()):
            return all_upcoming
        for upcoming_path in all_upcoming.values():
            upcoming_path.unlink(missing_ok=True)
        return {}

    @cached_property
    def latest_upcoming_mtime(self) -> float:
        'Return the latest mtime of the upcoming certificate.'
        return max((
            Path.stat(x).st_mtime
            for x in self.all_upcomings_in_place.values()
        ), default=0)

    @cached_property
    def common_names(self) -> list[str]:
        'The list of hostnames this certificate represents.'
        return self.names()

    def set_upcoming_as_latest(self) -> None:
        """
        Put the certificate from the upcoming directory into the
        archive, and link it as the current (latest) certificate.
        """
        archive_path = Path(self.archive_dir)
        for pem_path in self.all_upcomings_in_place.values():
            new_path = archive_path.joinpath(pem_path.name)
            if pem_path.is_symlink():
                # It's a private key symlinked to a former one (pinned)
                points_at = pem_path.readlink().absolute()
                pem_path.unlink()
                new_path.symlink_to(target=new_path.get_relative(to=points_at))
                continue
            pem_path.rename(new_path)
            # Modify mtime to know when to remove old TLSA records
            new_path.touch(exist_ok=True)
        self.update_all_links_to(version=self.latest_common_version())
        self.__reset_cache()

    def __handle_privkey_save(
        self, new_privkey: Key, old_privkey: Path, new_path: Path,
        is_privkey_reused: bool
    ) -> None:
        """
        Save a newly fetched private key, or link to its original.

        See `certbot._internal.storage.RenewableCert.save_successor()`.
        """
        if is_privkey_reused:
            if old_privkey.is_symlink():
                old_privkey = old_privkey.readlink()
            relative_path = old_privkey.relative_to(new_path)
            _LOGGER.debug(
                msg=f'Writing symlink to old private key, {relative_path}.')
            new_path.symlink_to(target=relative_path)
            return
        new_path.write_bytes(data=new_privkey.pem)
        # Preserve gid and (mode & MASK_FOR_PRIVATE_KEY_PERMISSIONS)
        # from previous privkey in this lineage.
        str_old_privkey = str(old_privkey)
        mode = compute_private_key_mode(
            old_key=str_old_privkey, base_mode=BASE_PRIVKEY_MODE)
        copy_ownership_and_apply_mode(
            src=str_old_privkey, dst=str(new_path), mode=mode, copy_user=False,
            copy_group=True)

    def __handle_pubkey_save(
        self, targets: _TargetDictType, new_cert: bytes, new_chain: bytes
    ) -> None:
        """
        Save a newly fetched public key and chain/fullchain.

        See `certbot._internal.storage.RenewableCert.save_successor()`.
        """
        _LOGGER.debug(msg=f'Writing certificate to {targets["cert"]}')
        targets['cert'].write_bytes(data=new_cert)
        _LOGGER.debug(msg=f'Writing chain to {targets["chain"]}')
        targets['chain'].write_bytes(data=new_chain)
        _LOGGER.debug(msg=f'Writing full chain to {targets["fullchain"]}')
        targets['fullchain'].write_bytes(data=new_cert + new_chain)

    def __update_configfile(self):
        """
        I guess this is for when cert paths or `ALL_FOUR` change.

        See `certbot._internal.storage.RenewableCert.save_successor()`.
        """
        symlinks: Mapping[str, str] = {
            kind: self.configuration[kind] for kind in ALL_FOUR
        }  # type: ignore
        # Update renewal config file
        self.configfile = config = update_configuration(
            lineagename=self.lineagename,
            archive_dir=self.archive_dir, target=symlinks,
            cli_config=self.cli_config)
        self.configuration = \
            config_with_defaults(config=config)  # type: ignore

    def save_to_upcoming(
        self, new_cert: bytes, new_chain: bytes, new_privkey: Key,
        is_privkey_reused: bool
    ) -> None:
        """
        Put the new certificate into the upcoming directory.

        Original save logic:
        :func:`certbot._internal.storage.RenewableCert.save_successor`
        """
        prior_version = self.latest_common_version()
        new_version = self.next_free_version()
        targets: _TargetDictType = {
            kind: self.upcoming_dir.joinpath(f'{kind}{new_version}.pem')
            for kind in ALL_FOUR}
        old_privkey = Path(self.archive_dir).joinpath(
            f'privkey{prior_version}.pem')
        self.__handle_privkey_save(
            new_privkey=new_privkey, old_privkey=old_privkey,
            new_path=targets['privkey'], is_privkey_reused=is_privkey_reused)
        self.__handle_pubkey_save(
            targets=targets, new_cert=new_cert, new_chain=new_chain)
        self.__update_configfile()
        self.__reset_cache()

    @cached_property
    def __latest_pubcert(self) -> Certificate:
        data = Path(self.cert_path).read_bytes()
        return load_pem_x509_certificate(data=data)

    @cached_property
    def __upcoming_pubcert(self) -> Optional[Certificate]:
        if not self.all_upcomings_in_place:
            return
        data = Path(self.all_upcomings_in_place['cert']).read_bytes()
        return load_pem_x509_certificate(data=data)

    @cached_property
    def __second_latest_pubcert(self) -> Certificate:
        'Return the second-latest public `Certificate`.'
        version = self.latest_common_version() - 1
        cert_path = Path(self.version(kind='cert', version=version))
        data = Path(cert_path).read_bytes()
        return load_pem_x509_certificate(data=data)

    @lru_cache(maxsize=None)
    def __is_oldcert_tlsa_needed(self, time_to_reload: int) -> bool:
        """
        Return `True` if the old certificate's TLSA records are needed.
        """
        if self.all_upcomings_in_place:
            # Upcoming cert in place
            return False
        utcnow = datetime.now(tz=timezone.utc).timestamp()
        return self.latest_installed_mtime + time_to_reload > utcnow

    @lru_cache(maxsize=None)
    def get_tlsa_recordtypes(
            self, usage: TlsaUsage, selector: TlsaSelector,
            matching_type: TlsaMatchingType, ttl: int,
            time_to_reload: int) -> list[TlsaResourceRecord]:
        'Return a list of generated `TlsaRecordType`s.'
        result = list[TlsaResourceRecord]()
        data = get_tlsa_bytes(
            cert=self.__latest_pubcert, selector=selector,
            matching_type=matching_type)
        result.append(TlsaResourceRecord(
            ttl=ttl, usage=usage, selector=selector,
            matching_type=matching_type, data=data))
        if self.__upcoming_pubcert:
            data = get_tlsa_bytes(
                cert=self.__upcoming_pubcert, selector=selector,
                matching_type=matching_type)
            result.append(TlsaResourceRecord(
                ttl=ttl, usage=usage, selector=selector,
                matching_type=matching_type, data=data))
        if self.__is_oldcert_tlsa_needed(time_to_reload=time_to_reload):
            data = get_tlsa_bytes(
                cert=self.__second_latest_pubcert, selector=selector,
                matching_type=matching_type)
            result.append(TlsaResourceRecord(
                ttl=ttl, usage=usage, selector=selector,
                matching_type=matching_type, data=data))
        return result
