from __future__ import annotations

from abc import ABCMeta
from collections.abc import Generator
from dataclasses import dataclass
from functools import cached_property
from io import TextIOWrapper
from ipaddress import IPv4Address, IPv6Address, ip_address
from os import W_OK, access
from pathlib import Path
from pprint import pformat
from typing import Any, Literal, Optional, Union, get_origin

from dns.name import Name
from dns.tsig import Key
from dns.tsigkeyring import from_text as tsig_from_text
from typing_extensions import Self, TypedDict, is_typeddict

from .config_types.common import (
    LITERAL_TCP_OR_UDP, LITERAL_TSIG_ALGORITHMS, TlsaMatchingType,
    TlsaProtocol, TlsaSelector, TlsaUsage)

try:
    from yaml import CDumper as Dumper
    from yaml import CLoader as Loader
except ImportError:
    from yaml import Dumper, Loader

from yaml import dump, load

from .config_types.configyaml import (
    ConfDict, DefaultsDict, DefaultsTlsaDict, DefaultsTlsaPortDict,
    DefaultTlsaRrtypeDict, DnsUpdateDict, DnsUpdateServerDict,
    DnsUpdateServerTsigDict, PathsDict, PathsNsupdateDict)
from .config_types.runtimeyaml import (
    RUNTIMECONF_DOC, RUNTIMECONF_TEMPLATE, RuntimeAdopteditemDict,
    RuntimeAdopteditemHostDict, RuntimeAdopteditemHostitemRecordsitemDict,
    RuntimeAdopteditemHostitemRecordsitemRrtypeitemDict,
    RuntimeAdopteditemServeritemDict, RuntimeAdopteditemServeritemTsigDict,
    RuntimeDict)


def get_host_and_domain(hostname: str) -> tuple[str, str]:
    'Return a guessed host and zone from a hostname.'
    splitted = hostname.split(sep='.')
    if len(splitted) == 2:
        # Root of a zone
        return '.', hostname
    return '.'.join(splitted[:-2]), '.'.join(splitted[-2:])


def check_required_keys(data: Any, datatype: ABCMeta) -> None:
    """
    Check if all the required keys are set in the passed `datatype` (a
    TypedDict).
    """
    if not isinstance(data, dict):
        raise RuntimeError(
            f'Passed data is not a dict instance: {type(data)}, {data!r}')
    if not is_typeddict(tp=datatype):
        raise RuntimeError(
            f'Passed datatype is not a TypedDict type: {datatype}')
    casted_datatype: TypedDict = datatype  # type: ignore
    missing_keys = casted_datatype.__required_keys__ - set(data)
    if missing_keys:
        raise RuntimeError(f'Missing keys from {datatype}: {missing_keys}')
    for key in casted_datatype.__required_keys__:
        expected_type = datatype.__annotations__[key]
        type_origin = get_origin(expected_type)
        if is_typeddict(expected_type):
            expected_type = dict
        elif type_origin is Literal:
            expected_type = str
        elif type_origin:
            expected_type = type_origin
        value = data[key]
        if not isinstance(value, expected_type):
            expected_type = datatype.__annotations__[key]
            raise RuntimeError(
                f'{key!r} in {type(data)} is not {expected_type}')


def _check_write_access(path: Path) -> None:
    'Raise if the path is not writeable.'
    if not access(path=path, mode=W_OK):
        raise RuntimeError(f'{path} is not writeable')


@dataclass
class _DefaultsTlsaPortConf(object):
    protocol: LITERAL_TCP_OR_UDP
    portnumber: int

    @staticmethod
    def parse(data: DefaultsTlsaPortDict) -> _DefaultsTlsaPortConf:
        check_required_keys(data=data, datatype=DefaultsTlsaPortDict)
        return _DefaultsTlsaPortConf(
            protocol=data['protocol'], portnumber=data['portnumber'])


@dataclass
class DefaultsTlsaRrtypeConf(object):
    usage: TlsaUsage
    selector: TlsaSelector
    matching_type: TlsaMatchingType
    _singletons = dict[tuple[int, int, int], Self]()

    @staticmethod
    def parse(data: DefaultTlsaRrtypeDict) -> DefaultsTlsaRrtypeConf:
        key = (data['usage'], data['selector'], data['matching_type'])
        if result := DefaultsTlsaRrtypeConf._singletons.get(key):
            return result
        result = DefaultsTlsaRrtypeConf(
            usage=TlsaUsage(value=data['usage']),
            selector=TlsaSelector(value=data['selector']),
            matching_type=TlsaMatchingType(value=data['matching_type']))
        DefaultsTlsaRrtypeConf._singletons[key] = result
        return result

    @cached_property
    def adopted_type(self) -> AdoptedCertHostitemRecordsitemRrtypeitemConfig:
        return AdoptedCertHostitemRecordsitemRrtypeitemConfig(
            usage=self.usage, selector=self.selector,
            matching_type=self.matching_type)


@dataclass
class _DefaultsTlsaConf(object):
    ttl: int
    time_to_reload: int
    records: list[_DefaultsTlsaPortConf]
    rrtypes: list[DefaultsTlsaRrtypeConf]

    @staticmethod
    def parse(data: DefaultsTlsaDict) -> _DefaultsTlsaConf:
        'Return a parsed version from yaml data.'
        check_required_keys(data=data, datatype=DefaultsTlsaDict)
        return _DefaultsTlsaConf(
            ttl=data['ttl'], time_to_reload=data['time_to_reload'],
            records=[
                _DefaultsTlsaPortConf.parse(data=x) for x in data['records']],
            rrtypes=[
                DefaultsTlsaRrtypeConf.parse(data=x) for x in data['rrtypes']]
        )


@dataclass
class _DefaultsConf(object):
    '`defaults` section of the configuration.'
    tlsa: _DefaultsTlsaConf
    method: Literal['dns-update']

    @staticmethod
    def parse(data: DefaultsDict) -> _DefaultsConf:
        'Return a parsed version from yaml data.'
        check_required_keys(data=data, datatype=DefaultsDict)
        return _DefaultsConf(
            tlsa=_DefaultsTlsaConf.parse(data=data['tlsa']),
            method=data['method'])


@dataclass
class _DnsUpdateServerTsigConf(object):
    name: str
    algorithm: LITERAL_TSIG_ALGORITHMS
    secret: str

    @staticmethod
    def parse(data: DnsUpdateServerTsigDict) -> _DnsUpdateServerTsigConf:
        return _DnsUpdateServerTsigConf(
            name=data['name'], algorithm=data['algorithm'],
            secret=data['secret'])

    @cached_property
    def keyring(self) -> dict[Name, Key]:
        'A transmission formatted keyring.'
        return tsig_from_text(
            textring={self.name: (self.algorithm, self.secret)})


@dataclass
class _DnsUpdateServerConf(object):
    ip: Union[IPv4Address, IPv6Address]
    tsig: _DnsUpdateServerTsigConf

    @staticmethod
    def parse(data: DnsUpdateServerDict) -> _DnsUpdateServerConf:
        check_required_keys(data=data, datatype=DnsUpdateServerDict)
        ip = ip_address(address=data['ip'])
        tsig = _DnsUpdateServerTsigConf.parse(data=data['tsig'])
        return _DnsUpdateServerConf(ip=ip, tsig=tsig)


@dataclass
class _DnsServerInfo(object):
    'Collected information about one DNS server.'
    ip: Union[IPv4Address, IPv6Address]
    tsig: Union[_DnsUpdateServerTsigConf, AdoptedCertServeritemTsigConfig]


@dataclass
class _DnsUpdateConf(object):
    '`dns-update` section of the configuration.'
    servers: dict[Union[IPv4Address, IPv6Address], _DnsUpdateServerConf]

    @staticmethod
    def parse(data: DnsUpdateDict) -> _DnsUpdateConf:
        'Return a parsed version from yaml data.'
        check_required_keys(data=data, datatype=DnsUpdateDict)
        servers = dict[Union[IPv4Address, IPv6Address], _DnsUpdateServerConf]()
        for item in data['servers']:
            parsed = _DnsUpdateServerConf.parse(data=item)
            servers[parsed.ip] = parsed
        return _DnsUpdateConf(servers=servers)

    @cached_property
    def serverupdates(self) -> list[_DnsServerInfo]:
        'A list of `_ServerUpdate`s from config.yaml values'
        return [
            _DnsServerInfo(ip=value.ip, tsig=value.tsig)
            for value in self.servers.values()]


@dataclass
class _PathsNsupdateConf(object):
    '`paths["nsupdate"]` section of the configuration.'
    executable: str
    params: list[str]

    @staticmethod
    def parse(data: PathsNsupdateDict) -> _PathsNsupdateConf:
        'Return a parsed version from yaml data.'
        check_required_keys(data=data, datatype=PathsNsupdateDict)
        return _PathsNsupdateConf(
            executable=data['executable'], params=data['params'])


@dataclass
class _PathsConf(object):
    '`paths` section of the configuration.'
    runtimeyaml: Path
    nsupdate: _PathsNsupdateConf

    @staticmethod
    def parse(data: PathsDict) -> _PathsConf:
        'Return a parsed version from yaml data.'
        check_required_keys(data=data, datatype=PathsDict)
        runtimeyaml = Path(data['runtimeyaml'])
        if runtimeyaml.exists():
            # See if file/dir is writeable
            _check_write_access(path=runtimeyaml)
            if runtimeyaml.is_dir():
                runtimeyaml = runtimeyaml.joinpath('daneupdate-runtime.yaml')
        else:
            # See if parent dir is writeable
            parent_path = runtimeyaml.parent
            parent_path.mkdir(parents=True, exist_ok=True)
            _check_write_access(path=parent_path)
        return _PathsConf(
            runtimeyaml=runtimeyaml,
            nsupdate=_PathsNsupdateConf.parse(data=data['nsupdate']))


@dataclass
class AdoptedCertServeritemTsigConfig(object):
    name: str
    algorithm: LITERAL_TSIG_ALGORITHMS
    secret: str

    @staticmethod
    def parse(
        item: RuntimeAdopteditemServeritemTsigDict
    ) -> AdoptedCertServeritemTsigConfig:
        'Return a parsed `AdoptedCertServeritemTsigConfig`.'
        check_required_keys(
            data=item, datatype=RuntimeAdopteditemServeritemTsigDict)
        return AdoptedCertServeritemTsigConfig(
            name=item['name'], algorithm=item['algorithm'],
            secret=item['secret'])

    def dump(self) -> RuntimeAdopteditemServeritemTsigDict:
        'Return a dumped `RuntimeAdopteditemServeritemTsigDict`.'
        return RuntimeAdopteditemServeritemTsigDict(
            name=self.name, algorithm=self.algorithm, secret=self.secret)

    @cached_property
    def keyring(self) -> dict[Name, Key]:
        'A transmission formatted keyring.'
        return tsig_from_text(
            textring={self.name: (self.algorithm, self.secret)})


@dataclass
class AdoptedCertHostitemServeritemConfig(object):
    'One server configuration for one adopted certificate.'
    ip: Union[IPv4Address, IPv6Address]
    tsig: Optional[AdoptedCertServeritemTsigConfig]

    @staticmethod
    def parse(
        item: RuntimeAdopteditemServeritemDict
    ) -> AdoptedCertHostitemServeritemConfig:
        'Return a parsed `AdoptedCertHostitemServeritemConfig`.'
        check_required_keys(
            data=item, datatype=RuntimeAdopteditemServeritemDict)
        tsig = item.get('tsig')
        if tsig is not None:
            tsig = AdoptedCertServeritemTsigConfig.parse(item=tsig)
        return AdoptedCertHostitemServeritemConfig(
            ip=ip_address(address=item['ip']), tsig=tsig)

    def dump(self) -> RuntimeAdopteditemServeritemDict:
        'Return a dumped `RuntimeAdopteditemServeritemDict`.'
        if not self.tsig:
            return RuntimeAdopteditemServeritemDict(ip=str(self.ip))
        return RuntimeAdopteditemServeritemDict(
            ip=str(self.ip), tsig=self.tsig.dump())

    def get_serverinfo(self, config: Configuration) -> _DnsServerInfo:
        'Return a `_ServerUpdate` with all variables filled.'
        if self.tsig:
            return _DnsServerInfo(ip=self.ip, tsig=self.tsig)
        if item := config.dnsupdate.servers.get(self.ip):
            return _DnsServerInfo(ip=self.ip, tsig=item.tsig)
        raise KeyError(
            f'No {self.ip} found in config["dns-update"]["servers"]: ' +
            pformat(list(config.dnsupdate.servers.values())))


@dataclass
class AdoptedCertHostitemRecordsitemRrtypeitemConfig(object):
    usage: TlsaUsage
    selector: TlsaSelector
    matching_type: TlsaMatchingType

    @staticmethod
    def parse(
        item: RuntimeAdopteditemHostitemRecordsitemRrtypeitemDict
    ) -> AdoptedCertHostitemRecordsitemRrtypeitemConfig:
        return AdoptedCertHostitemRecordsitemRrtypeitemConfig(
            usage=TlsaUsage(value=item['usage']),
            selector=TlsaSelector(value=item['selector']),
            matching_type=TlsaMatchingType(value=item['matching_type']))

    def dump(self) -> RuntimeAdopteditemHostitemRecordsitemRrtypeitemDict:
        return RuntimeAdopteditemHostitemRecordsitemRrtypeitemDict(
            usage=self.usage.value, selector=self.selector.value,
            matching_type=self.matching_type.value)


@dataclass
class AdoptedCertHostitemRecordsitemConfig(object):
    protocol: TlsaProtocol
    portnumber: int
    ttl: Optional[int]
    time_to_reload: Optional[int]
    rrtypes: Optional[list[AdoptedCertHostitemRecordsitemRrtypeitemConfig]]

    @staticmethod
    def parse(
        item: RuntimeAdopteditemHostitemRecordsitemDict
    ) -> AdoptedCertHostitemRecordsitemConfig:
        'Return a parsed `AdoptedCertHostitemPortsitemConfig`.'
        check_required_keys(
            data=item, datatype=RuntimeAdopteditemHostitemRecordsitemDict)
        usage = item.get('usage')
        if isinstance(usage, list):
            usage = TlsaUsage.from_ints(data=usage)
        selector = item.get('selector')
        if isinstance(selector, list):
            selector = TlsaSelector.from_ints(data=selector)
        matching_type = item.get('matching_type')
        if isinstance(matching_type, list):
            matching_type = TlsaMatchingType.from_ints(data=matching_type)
        rrtypes = item.get('rrtypes')
        if isinstance(rrtypes, list):
            rrtypes = [
                AdoptedCertHostitemRecordsitemRrtypeitemConfig.parse(item=x)
                for x in rrtypes]
        return AdoptedCertHostitemRecordsitemConfig(
            protocol=TlsaProtocol.from_config_literal(data=item['protocol']),
            portnumber=item['portnumber'], ttl=item.get('ttl'),
            time_to_reload=item.get('time_to_reload'), rrtypes=rrtypes)

    def dump(self) -> RuntimeAdopteditemHostitemRecordsitemDict:
        'Return a dumped `AdoptedCertHostitemPortsitemConfig`.'
        result = RuntimeAdopteditemHostitemRecordsitemDict(
            protocol=self.protocol.to_literal(), portnumber=self.portnumber)
        if isinstance(self.ttl, int):
            result['ttl'] = self.ttl
        if isinstance(self.time_to_reload, int):
            result['time_to_reload'] = self.time_to_reload
        if self.rrtypes:
            result['rrtypes'] = [x.dump() for x in self.rrtypes]
        return result


@dataclass
class AdoptedCertHostitemConfig(object):
    'One host configuration for an adopted certificate.'
    name: str
    zone: Optional[str]
    records: Optional[list[AdoptedCertHostitemRecordsitemConfig]]
    servers: Optional[list[AdoptedCertHostitemServeritemConfig]]

    @staticmethod
    def parse(item: RuntimeAdopteditemHostDict) -> AdoptedCertHostitemConfig:
        'Return a parsed `AdoptedCertHostitemConfig`.'
        check_required_keys(data=item, datatype=RuntimeAdopteditemHostDict)
        servers = item.get('servers')
        if servers is not None:
            servers = [
                AdoptedCertHostitemServeritemConfig.parse(item=x)
                for x in servers]
        records = item.get('records')
        if records is not None:
            records = [
                AdoptedCertHostitemRecordsitemConfig.parse(item=x)
                for x in records]
        return AdoptedCertHostitemConfig(
            name=item['name'], zone=item.get('zone'), records=records,
            servers=servers)

    def get_serverinfo(self, config: Configuration) -> list[_DnsServerInfo]:
        'Return a compiled `_ServerUpdate` list.'
        if self.servers is None:
            return config.dnsupdate.serverupdates
        return [item.get_serverinfo(config=config) for item in self.servers]

    def dump(self) -> RuntimeAdopteditemHostDict:
        'Return a dumped `RuntimeAdopteditemHostDict`.'
        result = RuntimeAdopteditemHostDict(name=self.name)
        if isinstance(self.zone, str):
            result['zone'] = self.zone
        if isinstance(self.records, list):
            result['records'] = [x.dump() for x in self.records]
        if isinstance(self.servers, list):
            result['servers'] = [x.dump() for x in self.servers]
        return result


@dataclass
class AdoptedCertConfig(object):
    'One adopted certificate configuration.'
    configpath: str
    hosts: Optional[list[AdoptedCertHostitemConfig]]

    @staticmethod
    def parse(item: RuntimeAdopteditemDict) -> AdoptedCertConfig:
        'Return a parsed `AdoptedCertConfig`.'
        check_required_keys(data=item, datatype=RuntimeAdopteditemDict)
        hosts = item.get('hosts')
        if isinstance(hosts, list):
            hosts = [AdoptedCertHostitemConfig.parse(item=x) for x in hosts]
        return AdoptedCertConfig(configpath=item['configpath'], hosts=hosts)

    @staticmethod
    def get_empty(configpath: str) -> AdoptedCertConfig:
        'Return a minimal variant of an `AdoptedCertConfig`.'
        return AdoptedCertConfig(configpath=configpath, hosts=None)

    def dump(self) -> RuntimeAdopteditemDict:
        'Return a dumped `RuntimeAdopteditemDict`.'
        result = RuntimeAdopteditemDict(configpath=self.configpath)
        if self.hosts is not None:
            result['hosts'] = [x.dump() for x in self.hosts]
        return result

    @cached_property
    def max_ttl(self) -> int:
        'Return the biggest TTL in the records or `0` if there is none.'
        result = set([0])
        for host in self.hosts or []:
            for record in host.records or []:
                if isinstance(record.ttl, int):
                    result.add(record.ttl)
        return max(result)


@dataclass
class _RuntimeConfiguration(object):
    'Runtime configuration that changes.'
    # `adopted`'s: lineage filepath (config path) -> AdoptedCertConfig
    adopted: dict[str, AdoptedCertConfig]
    config_filename = 'daneupdate-runtime.yaml'

    @staticmethod
    def load(path: Path) -> _RuntimeConfiguration:
        'Return a loaded `_RuntimeConfiguration`.'
        if not path.exists():
            path.write_text(data=RUNTIMECONF_DOC + RUNTIMECONF_TEMPLATE)
        with path.open() as fd:
            data: RuntimeDict = load(stream=fd, Loader=Loader)
        check_required_keys(data=data, datatype=RuntimeDict)
        adopted = dict[str, AdoptedCertConfig]()
        for item in data.get('adopted', []):
            adopted[item['configpath']] = AdoptedCertConfig.parse(item=item)
        return _RuntimeConfiguration(adopted=adopted)

    def save(self, path: Path) -> None:
        'Save the configuration back to YAML format.'
        data = RuntimeDict(adopted=[x.dump() for x in self.adopted.values()])
        with path.open(mode='w') as fd:
            fd.write(RUNTIMECONF_DOC)
            dump(data=data, stream=fd, Dumper=Dumper)


@dataclass
class Configuration(object):
    'Depicting a loaded configuration.'
    defaults: _DefaultsConf
    dnsupdate: _DnsUpdateConf
    paths: _PathsConf
    runtime: _RuntimeConfiguration

    @staticmethod
    def parse(wrapper: TextIOWrapper) -> Configuration:
        'Return the parsed configuration from the passed file.'
        with wrapper as fd:
            data: ConfDict = load(stream=fd, Loader=Loader)
        check_required_keys(data=data, datatype=ConfDict)
        paths = _PathsConf.parse(data=data['paths'])
        runtime = _RuntimeConfiguration.load(path=paths.runtimeyaml)
        return Configuration(
            defaults=_DefaultsConf.parse(data=data['defaults']),
            dnsupdate=_DnsUpdateConf.parse(data=data['dns-update']),
            runtime=runtime, paths=paths)

    def __hash__(self) -> int:
        return id(self)
