from __future__ import annotations

from enum import Enum, IntEnum
from typing import Literal, TypeVar

LITERAL_TCP_OR_UDP = Literal['tcp', 'udp']
LITERAL_TSIG_ALGORITHMS = Literal['hmac-sha256', 'hmac-sha512']


class _TlsaRecordBase(IntEnum):

    @classmethod
    def from_ints(cls: type[_T], data: list[int]) -> list[_T]:
        return [cls(value=x) for x in data]


_T = TypeVar('_T', bound=_TlsaRecordBase)


class TlsaUsage(_TlsaRecordBase):
    """
    Certificate usage.

    For e-mail (SMTP) applications, this must be either 2 (DANE-TA) or 3
    (DANE-EE). Values usage=0 and usage=1 should not be used for SMTP
    see rfc7672 section 3.1.3:
    https://www.rfc-editor.org/rfc/rfc7672.html#section-3.1.3
    """

    # A value of 0 is for what is commonly called CA constraint
    # (and PKIX-TA). The certificate provided when establishing TLS must
    # be issued by the listed root-CA or one of its intermediate CAs,
    # with a valid certification path to a root-CA already trusted by
    # the application doing the verification. The record may just point
    # to an intermediate CA, in which case the certificate for this
    # service must come via this CA, but the entire chain to a trusted
    # root-CA must still be valid.
    PKIX_TA = 0

    # A value of 1 is for what is commonly called service certificate
    # constraint (and PKIX-EE). The certificate used must match the TLSA
    # record, and it must also pass PKIX certification path validation
    # to a trusted root-CA.
    PKIX_EE = 1

    # A value of 2 is for what is commonly called trust anchor assertion
    # (and DANE-TA). The TLSA record matches the certificate of the root
    # CA, or one of the intermediate CAs, of the certificate in use by
    # the service. The certification path must be valid up to the
    # matching certificate, but there is no need for a trusted root-CA.
    DANE_TA = 2

    # A value of 3 is for what is commonly called domain issued
    # certificate (and DANE-EE). The TLSA record matches the used
    # certificate itself. The used certificate does not need to be
    # signed by other parties. This is useful for self-signed
    # certificates, but also for cases where the validator does not have
    # a list of trusted root certificates.
    DANE_EE = 3


class TlsaSelector(_TlsaRecordBase):
    """
    Certificate selector. See:
    https://en.wikipedia.org/wiki/DNS-based_Authentication_of_Named_Entities
    """

    # A value of 0 means to select the entire certificate for matching.
    ENTIRE = 0

    # A value of 1 means to select just the public key for certificate
    # matching. Matching the public key is often sufficient, as this is
    # likely to be unique.
    PUBLICKEY = 1


class TlsaMatchingType(_TlsaRecordBase):
    """
    Certificate matching type. See:
    https://en.wikipedia.org/wiki/DNS-based_Authentication_of_Named_Entities
    """

    # A type of 0 means the entire information selected is present in
    # the certificate association data.
    ENTIRE = 0

    # A type of 1 means to do a SHA-256 hash of the selected data.
    SHA256 = 1

    # A type of 2 means to do a SHA-512 hash of the selected data.
    SHA512 = 2


class TlsaProtocol(Enum):
    'Protocol of a TLSA record.'
    TCP = '_tcp'
    UDP = '_udp'

    @staticmethod
    def from_config_literal(data: LITERAL_TCP_OR_UDP) -> TlsaProtocol:
        'Return a `TlsaProtocol` loaded from a config value.'
        if data == 'tcp':
            return TlsaProtocol.TCP
        elif data == 'udp':
            return TlsaProtocol.UDP

    def to_literal(self) -> LITERAL_TCP_OR_UDP:
        'Return the config assignable value.'
        if self == TlsaProtocol.TCP:
            return 'tcp'
        elif self == TlsaProtocol.UDP:
            return 'udp'
        raise KeyError(f'{self} is invalid')
