from typing_extensions import NotRequired, TypedDict

from .common import LITERAL_TCP_OR_UDP, LITERAL_TSIG_ALGORITHMS

RUNTIMECONF_DOC = """\
# Fields — except for the configpath field — aren't mandatory, as they
# get their defaults from the configuration, or daneupdate will try to
# guess their values (e.g. the zone from the certificate name:
# x.example.com -> example.com).
#
# If adopted["hosts"] is specified, the `name` is the only mandatory
# parameter there, everything else is computed from the default
# configuration and certificate values.
#
# adopted:
# - configpath: test1.conf
#   # The root of the zone is marked as `.`
#   hosts:
#   - name: .
#     zone: example.com
#     records:
#     - protocol: tcp
#       portnumber: 443
#       ttl: 123
#       time_to_reload: 456
#       rrtypes:
#       - usage: 3
#         selector: 1
#         matching_type: 1
#       - usage: 3
#         selector: 1
#         matching_type: 2
#     - protocol: tcp
#       portnumber: 25
#       ttl: 123
#       time_to_reload: 456
#       rrtypes:
#       - usage: 3
#         selector: 1
#         matching_type: 1
#       - usage: 3
#         selector: 1
#         matching_type: 2
#     servers:
#     - ip: 127.0.0.1
#       tsig:
#         name: ddns-key-ipv4
#         algorithm: hmac-sha256
#         secret: super-secret-key
#     - ip: ::1
#       tsig:
#         name: ddns-key-ipv6
#         algorithm: hmac-sha256
#         secret: super-secret-key
# - configpath: test1.conf
#   hosts:
#   - name: testhost
#     zone: example.com
#     records:
#     - protocol: tcp
#       portnumber: 443
#       ttl: 123
#       time_to_reload: 456
#       rrtypes:
#       - usage: 3
#         selector: 1
#         matching_type: 1
#       - usage: 3
#         selector: 1
#         matching_type: 2
#     - protocol: tcp
#       portnumber: 25
#       ttl: 123
#       time_to_reload: 456
#       rrtypes:
#       - usage: 3
#         selector: 1
#         matching_type: 1
#       - usage: 3
#         selector: 1
#         matching_type: 2
#     ttl: 789
#     time_to_reload: 123
#     servers:
#     - ip: 127.0.0.1
#       tsig:
#         name: ddns-key-ipv4
#         algorithm: hmac-sha256
#         secret: super-secret-key
#     - ip: ::1
#       tsig:
#         name: ddns-key-ipv6
#         algorithm: hmac-sha256
#         secret: super-secret-key
"""
RUNTIMECONF_TEMPLATE = """\
adopted: []
"""


class RuntimeAdopteditemServeritemTsigDict(TypedDict):
    """
    `adopted["servers"]["tsig"]` section item of the runtime
    configuration.
    """
    name: str
    algorithm: LITERAL_TSIG_ALGORITHMS
    secret: str


class RuntimeAdopteditemServeritemDict(TypedDict):
    """
    `adopted["hosts"]["servers"]` section item of the runtime
    configuration.
    """
    ip: str
    tsig: NotRequired[RuntimeAdopteditemServeritemTsigDict]


class RuntimeAdopteditemHostitemRecordsitemRrtypeitemDict(TypedDict):
    """
    `adopted["hosts"]["records"]["rrtypes"]` section item of
    the runtime configuration.
    """
    usage: int
    selector: int
    matching_type: int


class RuntimeAdopteditemHostitemRecordsitemDict(TypedDict):
    """
    `adopted["hosts"]["records"]` section item of the runtime
    configuration.
    """
    protocol: LITERAL_TCP_OR_UDP
    portnumber: int
    ttl: NotRequired[int]
    time_to_reload: NotRequired[int]
    rrtypes: NotRequired[
        list[RuntimeAdopteditemHostitemRecordsitemRrtypeitemDict]]


class RuntimeAdopteditemHostDict(TypedDict):
    """
    Depicting the runtime config\'s `adopted[\'hosts\']` section format.
    """
    name: str
    zone: NotRequired[str]
    records: NotRequired[list[RuntimeAdopteditemHostitemRecordsitemDict]]
    servers: NotRequired[list[RuntimeAdopteditemServeritemDict]]


class RuntimeAdopteditemDict(TypedDict):
    'Depicting the runtime config\'s `adopted` section format.'
    configpath: str
    hosts: NotRequired[list[RuntimeAdopteditemHostDict]]


class RuntimeDict(TypedDict):
    'Depicting the runtime config format.'
    adopted: list[RuntimeAdopteditemDict]
