'Set up the command line `ArgumentParser` here.'

from argparse import ArgumentParser, FileType, _SubParsersAction
from typing import TYPE_CHECKING

from ..handlers.adopt import AdoptHandler, ListAdoptablesHandler
from ..handlers.renew import HELP_RENEW_DESC, HELP_RENEW_DRYRUN, RenewHandler
from ..handlers.update_tlsa import HELP_UPDATETLSA_DRYRUN, UpdateTlsaHandler

if TYPE_CHECKING:
    _SPA = _SubParsersAction[ArgumentParser]
else:
    _SPA = _SubParsersAction


def _add_baseargs(parser: ArgumentParser):
    'Add basic command line arguments.'
    parser.add_argument(
        '-c', '--config', help='Path of the configuration yaml', required=True,
        type=FileType('r', encoding='utf-8'))


def _setup_subparser_updatetlsa(subparsers: _SPA):
    'Setup `list-adoptables` subparser.'
    update_tlsa_parser = subparsers.add_parser(
        name='update-tlsa',
        help='Update TLSA records',
        description='Update TLSA records for certificates')
    update_tlsa_parser.add_argument(
        '--dry-run', action='store_true', help=HELP_UPDATETLSA_DRYRUN)
    update_tlsa_parser.add_argument(
        '-v', '--verbose', action='store_true',
        help='Be more verbose about TLSA records')
    update_tlsa_parser.add_argument(
        action='extend', nargs='*',
        help='One or more certificate name (hostname)', metavar='hostname',
        dest='hostnames')
    update_tlsa_parser.set_defaults(func=UpdateTlsaHandler)


def _setup_subparser_listadoptables(subparsers: _SPA):
    'Setup `list-adoptables` subparser.'
    list_adoptables_parser = subparsers.add_parser(
        name='list-adoptables',
        help='List the currently adoptable certificates',
        description='List the currently adoptable certificates')
    list_adoptables_parser.set_defaults(func=ListAdoptablesHandler)


def _setup_subparser_adopt(subparsers: _SPA):
    'Setup `adopt` subparser.'
    adopt_parser = subparsers.add_parser(
        name='adopt',
        help='Adopt a certificate deployed by certbot',
        description='Adopt a certificate deployed by certbot')
    adopt_parser.add_argument(
        '-v', '--verbose', action='store_true',
        help='Be more verbose about the adoption process')
    adopt_parser.add_argument(
        '-s', '--skip-tlsa-update', action='store_true',
        help='Skip setting up TLSA record for the adopted certificate.')
    adopt_parser.add_argument(
        '--dry-run', action='store_true', help='Emulate TLSA update.')
    adopt_parser.add_argument(
        action='extend', nargs='+',
        help='One or more certificate name (hostname)', metavar='hostname',
        dest='hostnames')
    adopt_parser.set_defaults(func=AdoptHandler)


def _setup_subparser_renew(subparsers: _SPA):
    'Setup `renew` subparser.'
    renew_parser = subparsers.add_parser(
        name='renew',
        help='Renew adopted and non-adopted certificates',
        description=HELP_RENEW_DESC)
    renew_parser.add_argument(
        '-t', '--test-cert', action='store_true',
        help='Use staging API endpoints')
    renew_parser.add_argument(
        '-f', '--force-renewal', action='store_true',
        help='Force renewal of certificates. This might break your certs!')
    renew_parser.add_argument(
        '--dry-run', action='store_true', help=HELP_RENEW_DRYRUN)
    renew_parser.add_argument(
        '-v', '--verbose', action='store_true',
        help='Be more verbose about the renewal process')
    renew_parser.add_argument(
        action='extend', nargs='*',
        help='One or more certificate name (hostname)', metavar='hostname',
        dest='hostnames')
    renew_parser.set_defaults(func=RenewHandler)


def _setup_subparsers(parser: ArgumentParser):
    'Setup subcommand parsers.'
    subparsers = parser.add_subparsers(
        title='sub-commands',
        description='Please use one of the following sub-commands:',
        required=True)
    _setup_subparser_listadoptables(subparsers=subparsers)
    _setup_subparser_adopt(subparsers=subparsers)
    _setup_subparser_renew(subparsers=subparsers)
    _setup_subparser_updatetlsa(subparsers=subparsers)


def get_parser() -> ArgumentParser:
    'Return an `ArgumentParser` that is set up to handle commands.'
    parser = ArgumentParser(
        description='DANE records updater.', prog='daneupdate')
    _add_baseargs(parser=parser)
    _setup_subparsers(parser=parser)
    return parser
