from __future__ import annotations

from dataclasses import dataclass
from functools import cached_property
from ipaddress import IPv4Address, IPv6Address
from logging import getLogger
from typing import Optional, Union

from certbot.display.util import notify
from certbot.errors import ConfigurationError
from dns.name import Name
from dns.name import from_text as name_from_text
from dns.nameserver import Do53Nameserver
from dns.query import tcp
from dns.rcode import Rcode
from dns.rdataclass import RdataClass
from dns.rdataset import Rdataset
from dns.rdatatype import TLSA, RdataType
from dns.rdtypes.ANY.TLSA import TLSA as RdataTlsa
from dns.resolver import NXDOMAIN, Resolver
from dns.rrset import RRset
from dns.tsig import Key
from dns.tsigkeyring import from_text as tsig_from_text
from dns.update import UpdateMessage

from ..cert import DaneCert
from ..config import (
    AdoptedCertHostitemConfig, AdoptedCertHostitemRecordsitemConfig,
    AdoptedCertHostitemRecordsitemRrtypeitemConfig,
    AdoptedCertHostitemServeritemConfig, Configuration, get_host_and_domain)
from ..config_types.common import TlsaProtocol
from .record_type import TlsaResourceRecord

_LOGGER = getLogger(name=__name__)


@dataclass
class _DnsServerTsig(object):
    name: str
    algorithm: str
    secret: str

    @cached_property
    def keyring(self) -> dict[Name, Key]:
        'A transmission formatted keyring.'
        return tsig_from_text(
            textring={self.name: (self.algorithm, self.secret)})

    @staticmethod
    def from_data(name: str, algorithm: str, secret: str):
        return _DnsServerTsig(name=name, algorithm=algorithm, secret=secret)


@dataclass
class _DnsServerInfo(object):
    ip: Union[IPv4Address, IPv6Address]
    tsig: _DnsServerTsig

    @staticmethod
    def get_defaults(config: Configuration) -> list[_DnsServerInfo]:
        'Return configuration defaults.'
        if result := _serverinfo_cache.get(config):
            return result
        result = list[_DnsServerInfo]()
        for item in config.dnsupdate.servers.values():
            tsig = item.tsig
            result.append(_DnsServerInfo(
                ip=item.ip, tsig=_DnsServerTsig.from_data(
                    name=tsig.name, algorithm=tsig.algorithm,
                    secret=tsig.secret)))
        _serverinfo_cache[config] = result
        return result

    @staticmethod
    def from_adopted(
        servers: list[AdoptedCertHostitemServeritemConfig],
        config: Configuration
    ) -> list[_DnsServerInfo]:
        result = list[_DnsServerInfo]()
        for item in servers:
            tsig_in = item.tsig
            if tsig_in is None:
                if s_item := config.dnsupdate.servers.get(item.ip):
                    tsig_in = s_item.tsig
            if tsig_in is None:
                raise ConfigurationError(
                    f'No tsig configuration found for server {item.ip}')
            tsig = _DnsServerTsig.from_data(
                name=tsig_in.name, algorithm=tsig_in.algorithm,
                secret=tsig_in.secret)
            result.append(_DnsServerInfo(ip=item.ip, tsig=tsig))
        return result

    @cached_property
    def resolver(self) -> Resolver:
        'A cached resolver.'
        my_nameserver = Do53Nameserver(address=self.str_ip)
        resolver = Resolver(configure=False)
        resolver.nameservers = [my_nameserver]
        return resolver

    @cached_property
    def str_ip(self) -> str:
        return str(self.ip)

    def _resolve(self, fqdn: _FqdnHostname) -> set[TlsaResourceRecord]:
        'Return resolved `TlsaRecordType` for the passed FQDN.'
        response_set = set[TlsaResourceRecord]()
        try:
            answer = self.resolver.resolve(qname=fqdn.as_dnsname, rdtype=TLSA)
        except NXDOMAIN:
            return response_set
        item: RdataTlsa
        ttl = answer.rrset.ttl if isinstance(answer.rrset, RRset) else 0
        for item in answer:  # type: ignore
            record = TlsaResourceRecord.from_rdtype(item=item, ttl=ttl)
            response_set.add(record)
        return response_set

    def _log_results(
            self, fqdn: _FqdnHostname, rdset_adding: Rdataset,
            rdset_removing: Rdataset, verbose: bool) -> None:
        if not verbose:
            notify(msg=(
                f'Nameserver {self.str_ip}, {fqdn}: adding '
                f'{len(rdset_adding.items)} record(s), removing '
                f'{len(rdset_removing.items)} record(s).'))
            return
        notify(msg=f'--- {fqdn}:')
        if rdset_adding.items:
            notify(msg=f'-- Adding:\n{rdset_adding}')
        else:
            notify(msg='-- Adding: -')
        if rdset_removing.items:
            notify(msg=f'-- Removing:\n{rdset_removing}')
        else:
            notify(msg='-- Removing: -')

    def _get_zoneupdate(
        self, fqdn: _FqdnHostname, records_existing: set[TlsaResourceRecord],
        records_required: set[TlsaResourceRecord], verbose: bool
    ) -> UpdateMessage:
        """
        Return an `UpdateMessage` for updating this FQDN. The assumption
        here is that there are missing or superfluous TLSA records, so
        the FQDN needs to be adjusted.
        """
        update = UpdateMessage(
            zone=fqdn.zone_as_dnsname, keyring=self.tsig.keyring)
        # Superfluous records (to remove)
        rdset_removing = Rdataset(rdclass=RdataClass.IN, rdtype=RdataType.TLSA)
        for record in records_existing - records_required:
            rdset_removing.add(rd=record.rdata, ttl=record.ttl)
        if rdset_removing.items:
            update.delete(fqdn.as_dnsname, rdset_removing)
        # Missing records (to add)
        rdset_adding = Rdataset(rdclass=RdataClass.IN, rdtype=RdataType.TLSA)
        for record in records_required - records_existing:
            rdset_adding.add(rd=record.rdata, ttl=record.ttl)
        if rdset_adding.items:
            update.add(fqdn.as_dnsname, rdset_adding)
        self._log_results(
            fqdn=fqdn, rdset_adding=rdset_adding,
            rdset_removing=rdset_removing, verbose=verbose)
        return update

    def _are_records_same(
        self, fqdn: _FqdnHostname, records_existing: set[TlsaResourceRecord],
        records_required: set[TlsaResourceRecord], verbose: bool
    ) -> bool:
        'Return trueness of similarity while logging.'
        if records_existing != records_required:
            return False
        if verbose:
            notify(msg=f'--- {fqdn}: Records are the same, skipping update.')
            return True
        notify(msg=(
            f'Nameserver {self.str_ip}, {fqdn}: Records are the same, '
            'skipping update.'))
        return True

    def update(
        self, hostname_records: _TlsaRecordsDict, verbose: bool, dry_run: bool
    ) -> None:
        'Compare necessary and existing records, update when necessary.'
        if verbose:
            notify(msg=f'\n---- Nameserver {self.ip}')
        for fqdn, records_required in hostname_records.items():
            records_existing = self._resolve(fqdn=fqdn)
            is_same = self._are_records_same(
                fqdn=fqdn, records_existing=records_existing,
                records_required=records_required, verbose=verbose)
            if is_same:
                continue
            update = self._get_zoneupdate(
                fqdn=fqdn, records_existing=records_existing,
                records_required=records_required, verbose=verbose)
            if dry_run:
                continue
            response = tcp(q=update, where=self.str_ip)
            resp_rcode = response.rcode()
            if resp_rcode != Rcode.NOERROR:
                raise RuntimeError(
                    f'Update failed, return code {resp_rcode!r}')


_serverinfo_cache = dict[Configuration, list[_DnsServerInfo]]()


@dataclass
class _TlsaHostPrefix(object):
    'The TLSA prefix to a hostname.'
    protocol: TlsaProtocol
    portnumber: int

    @cached_property
    def as_str(self) -> str:
        return f'_{self.portnumber}.{self.protocol.value}'

    def __hash__(self) -> int:
        return hash(self.as_str)


@dataclass
class _TlsaRecordsConfig(object):
    """
    Gathered `runtime.yaml` and `config.yaml` TLSA record information
    for a hostname.
    """
    prefix: _TlsaHostPrefix
    ttl: int
    time_to_reload: int
    rrtypes: list[AdoptedCertHostitemRecordsitemRrtypeitemConfig]

    @staticmethod
    def from_adopted(
        records: list[AdoptedCertHostitemRecordsitemConfig],
        config: Configuration
    ) -> list[_TlsaRecordsConfig]:
        """
        Return prepared records configuration by merging from the
        adopted and the default configuration.
        """
        result = list[_TlsaRecordsConfig]()
        def_tlsa = config.defaults.tlsa
        for item in records:
            ttl = item.ttl if isinstance(item.ttl, int) else def_tlsa.ttl
            time_to_reload = item.time_to_reload \
                if isinstance(item.time_to_reload, int) \
                else def_tlsa.time_to_reload
            rrtypes = item.rrtypes if isinstance(item.rrtypes, list) else [
                x.adopted_type for x in def_tlsa.rrtypes]
            prefix = _TlsaHostPrefix(
                protocol=item.protocol, portnumber=item.portnumber)
            result.append(_TlsaRecordsConfig(
                prefix=prefix, ttl=ttl, time_to_reload=time_to_reload,
                rrtypes=rrtypes))
        return result

    @staticmethod
    def get_defaults(config: Configuration) -> list[_TlsaRecordsConfig]:
        'Return configuration defaults.'
        result = list[_TlsaRecordsConfig]()
        def_tlsa = config.defaults.tlsa
        for item in def_tlsa.records:
            prefix = _TlsaHostPrefix(
                protocol=TlsaProtocol.from_config_literal(data=item.protocol),
                portnumber=item.portnumber)
            result.append(_TlsaRecordsConfig(
                prefix=prefix, ttl=def_tlsa.ttl,
                time_to_reload=def_tlsa.time_to_reload,
                rrtypes=[x.adopted_type for x in def_tlsa.rrtypes]))
        return result

    def required_records_for_cert(
            self, cert: DaneCert) -> set[TlsaResourceRecord]:
        """
        Return the currently required `TlsaRecordType`s for the passed
        `DaneCert`.
        """
        result = set[TlsaResourceRecord]()
        for rrtype in self.rrtypes:
            result.update(cert.get_tlsa_recordtypes(
                usage=rrtype.usage, selector=rrtype.selector,
                matching_type=rrtype.matching_type,
                ttl=self.ttl, time_to_reload=self.time_to_reload))
        return result


@dataclass
class _FqdnHostname(object):
    prefix: _TlsaHostPrefix
    # Hostname without the domain name
    hostname: str
    # The DNS zone to update
    zone: str
    # The domain name in the FQDN, defaults to zone
    domain: str

    @staticmethod
    def construct(
        prefix: _TlsaHostPrefix, name: str, zone: str,
        domain: Optional[str] = None
    ) -> _FqdnHostname:
        return _FqdnHostname(
            prefix=prefix, hostname=name, zone=zone, domain=domain or zone)

    @cached_property
    def _as_fqdn(self) -> str:
        fqdn = self.prefix.as_str
        if self.hostname != '.':
            fqdn += f'.{self.hostname}'
        fqdn += f'.{self.domain}.'
        return fqdn

    @cached_property
    def zone_as_dnsname(self) -> Name:
        return name_from_text(text=self.zone)

    @cached_property
    def as_dnsname(self) -> Name:
        return name_from_text(text=self._as_fqdn)

    def __str__(self) -> str:
        return self._as_fqdn

    def __hash__(self) -> int:
        return hash(self._as_fqdn)


_TlsaRecordsDict = dict[_FqdnHostname, set[TlsaResourceRecord]]


@dataclass
class HostnameConfig(object):
    """
    Collected config information about one hostname in a certificate.
    It's and evaluated configuration from config and runtime.
    """
    name: str
    zone: str
    servers: list[_DnsServerInfo]
    records: list[_TlsaRecordsConfig]

    @staticmethod
    def from_hostname(hostname: str, config: Configuration) -> HostnameConfig:
        """
        Return one configuration with defaults from a passed
        fully-qualified-domain-name, when configured hosts don't exist
        in `runtime.yaml` for this certificate.
        """
        host, zone = get_host_and_domain(hostname=hostname)
        return HostnameConfig(
            name=host, zone=zone,
            servers=_DnsServerInfo.get_defaults(config=config),
            records=_TlsaRecordsConfig.get_defaults(config=config))

    @staticmethod
    def from_adoptedhosts(
        hosts: list[AdoptedCertHostitemConfig], config: Configuration,
        cert: DaneCert
    ) -> list[HostnameConfig]:
        """
        Return a list of `LiveupdateConfig` from adopted runtime
        configurations.
        """
        result = list[HostnameConfig]()
        def_servers = _DnsServerInfo.get_defaults(config=config)
        def_records = _TlsaRecordsConfig.get_defaults(config=config)
        for item in hosts:
            zone = item.zone or \
                get_host_and_domain(hostname=cert.common_names[0])[1]
            servers = def_servers if item.servers is None \
                else _DnsServerInfo.from_adopted(
                    servers=item.servers, config=config)
            records = def_records if item.records is None \
                else _TlsaRecordsConfig.from_adopted(
                    records=item.records, config=config)
            result.append(HostnameConfig(
                name=item.name, zone=zone, servers=servers, records=records))
        return result

    def get_tlsa_records(self, cert: DaneCert) -> _TlsaRecordsDict:
        'Return TLSA records for the configuration here.'
        result = _TlsaRecordsDict()
        for record in self.records:
            key = _FqdnHostname.construct(
                prefix=record.prefix, name=self.name, zone=self.zone)
            result[key] = record.required_records_for_cert(cert=cert)
        return result
