from __future__ import annotations

from dataclasses import dataclass
from functools import cached_property

from dns.rdataclass import RdataClass
from dns.rdatatype import RdataType
from dns.rdtypes.ANY.TLSA import TLSA

from ..config_types.common import TlsaMatchingType, TlsaSelector, TlsaUsage


@dataclass
class TlsaResourceRecord(object):
    """
    A generated TLSA resource record type. See:
    https://en.wikipedia.org/wiki/DNS-based_Authentication_of_Named_Entities
    """
    ttl: int
    usage: TlsaUsage
    selector: TlsaSelector
    matching_type: TlsaMatchingType
    data: bytes

    @staticmethod
    def from_rdtype(item: TLSA, ttl: int):
        return TlsaResourceRecord(
            ttl=ttl, usage=TlsaUsage(value=item.usage),
            selector=TlsaSelector(value=item.selector),
            matching_type=TlsaMatchingType(value=item.mtype), data=item.cert)

    @cached_property
    def rdata(self) -> TLSA:
        return TLSA(
            rdclass=RdataClass.IN, rdtype=RdataType.TLSA,
            usage=self.usage.value, selector=self.selector.value,
            mtype=self.matching_type.value, cert=self.data)

    def __eq__(self, other: object) -> bool:
        if isinstance(other, TlsaResourceRecord):
            return self.ttl == other.ttl \
                and self.usage == other.usage \
                and self.selector == other.selector \
                and self.matching_type == other.matching_type \
                and self.data == other.data
        return False

    def __hash__(self) -> int:
        return hash((self.usage, self.selector, self.matching_type, self.data))
